/* Copyright (c) 2012-2017 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "Database.h"
#include "version.h"

// Qt includes
#include <QDebug>
#include <QSqlQuery>
#include <QMessageBox>
#include <QDir>
#include <QSqlError>

namespace {

// The database version we're using
constexpr int DB_VERSION = 3;

// Default values
constexpr int DEFAULT_BOOGER_SCORE    = 21;
constexpr int DEFAULT_GAMES_PER_ROUND = 2;

// Database string literals
const QString DB_DB_VERSION          = QString::fromUtf8("DB_VERSION");
const QString DB_MUCKTURNIER_VERSION = QString::fromUtf8("MUCKTURNIER_VERSION");
const QString DB_TOURNAMENT_MODE     = QString::fromUtf8("TOURNAMENT_MODE");
const QString DB_BOOGER_SCORE        = QString::fromUtf8("BOOGER_SCORE");
const QString DB_GAMES_PER_ROUND     = QString::fromUtf8("GAMES_PER_ROUND");
const QString DB_FIXED_PAIRS         = QString::fromUtf8("FIXED_PAIRS");
const QString DB_SINGLE_PLAYERS      = QString::fromUtf8("SINGLE_PLAYERS");

const QString DB_CONNECTION_NAME = QString::fromUtf8("SQLiteDB");

}

Database::Database(QObject *parent) : QObject(parent)
{
}

Database::~Database()
{
    closeDatabaseConnection();
}

void Database::closeDatabaseConnection()
{
    if (m_db.isOpen()) {
        m_db.commit();
        m_db.close();
        m_db = QSqlDatabase::database();
        QSqlDatabase::removeDatabase(DB_CONNECTION_NAME);
    }
}

bool Database::connectDatabase(QString &dbFile)
{
    closeDatabaseConnection();

    m_db = QSqlDatabase::addDatabase(QString::fromUtf8("QSQLITE"), DB_CONNECTION_NAME);
    m_db.setDatabaseName(dbFile);

    if (! m_db.open()) {
        // This should not happen
        QMessageBox::critical(dynamic_cast<QWidget *>(parent()),
            tr("Datenbankverbindung fehlgeschlagen"),
            tr("Unerwarteter Fehler: Konnte keine Verbindung zur Datenbank „%1“ herstellen.").arg(
               QDir::toNativeSeparators(dbFile)),
            QMessageBox::Ok);
        return false;
    }

    return true;
}

bool Database::create(QString dbFile)
{
    if (! connectDatabase(dbFile)) {
        return false;
    }

    m_db.transaction();
    QSqlQuery query(m_db);
    query.exec(QString::fromUtf8("CREATE TABLE config(\n"
                                 "    key TEXT,\n"
                                 "    value TEXT\n"
                                 ")"));
    query.exec(QString::fromUtf8("CREATE TABLE players(\n"
                                 "    id INTEGER PRIMARY KEY,\n"
                                 "    name TEXT\n"
                                 ")"));
    query.exec(QString::fromUtf8("CREATE TABLE assignment(\n"
                                 "    round INTEGER,\n"
                                 "    pair_or_player INTEGER,\n"
                                 "    pair_number INTEGER,\n"
                                 "    table_number INTEGER\n"
                                 ")"));
    query.exec(QString::fromUtf8("CREATE TABLE score(\n"
                                 "    round INTEGER,\n"
                                 "    table_number INTEGER,\n"
                                 "    game INTEGER,\n"
                                 "    pair_or_player INTEGER,\n"
                                 "    score INTEGER\n"
                                 ")"));
    m_db.commit();

    setStandardValues();
    updateSettings();
    m_tournamentStarted = false;

    return true;
}

void Database::setStandardValues()
{
    m_db.transaction();
    QSqlQuery query(m_db);

    query.prepare(QString::fromUtf8("INSERT INTO config(key, value) VALUES(?, ?)"));

    QVariantList keys;
    keys << DB_MUCKTURNIER_VERSION
         << DB_DB_VERSION
         << DB_TOURNAMENT_MODE
         << DB_BOOGER_SCORE
         << DB_GAMES_PER_ROUND;
    query.addBindValue(keys);

    QVariantList values;
    values << QString::fromUtf8(MT_VERSION)
           << DB_VERSION
           << DB_FIXED_PAIRS
           << DEFAULT_BOOGER_SCORE
           << DEFAULT_GAMES_PER_ROUND;
    query.addBindValue(values);

    query.execBatch();
    m_db.commit();
}

void Database::updateSettings()
{
    // Reads the settings stored in the database.

    QSqlQuery query(m_db);
    query.exec(QString::fromUtf8("SELECT key, value FROM config"));

    while (query.next()) {
        QString key = query.value(0).toString();
        QString value = query.value(1).toString();

        if (key == DB_MUCKTURNIER_VERSION) {
            m_dbValueMuckturnierVersion = value;
        } else if (key == DB_TOURNAMENT_MODE) {
            if (value == DB_FIXED_PAIRS) {
                m_settingsTournamentMode = TournamentMode::fixedPairs;
            } else if (value == DB_SINGLE_PLAYERS) {
                m_settingsTournamentMode = TournamentMode::singlePlayers;
            }
        } else if (key == DB_DB_VERSION) {
            m_dbValueDbVersion = value.toInt();
        } else if (key == DB_BOOGER_SCORE) {
            m_settingsBoogerScore = value.toInt();
        } else if (key == DB_GAMES_PER_ROUND) {
            m_settingsGamesPerRound = value.toInt();
        }
    }
}

bool Database::tournamentStarted()
{
    return m_tournamentStarted;
}

bool Database::openDb(QString dbFile)
{
    m_openError.clear();

    if (! connectDatabase(dbFile)) {
        return false;
    }

    QSqlQuery query(m_db);

    // Let's see if we have all tables
    query.exec(QString::fromUtf8("SELECT name "
                                 "FROM sqlite_master "
                                 "WHERE "
                                     "type = 'table' "
                                     "AND name = 'config' "
                                     "OR name = 'players' "
                                     "OR name = 'assignment' "
                                     "OR name = 'score'"));
    int tableCount = 0;
    while (query.next()) {
        tableCount++;
    }
    if (tableCount != 4) {
        m_openError = tr("In der Datenbank sind nicht alle nötigen Tabllen vorhanden. "
                         "vermutlich ist „%1“ keine Muckturnier-Datenbank, oder die Datei wurde "
                         "mit einer zu alten Version von Muckturnier (vor 0.5.1) erstellt!").arg(
                         QDir::toNativeSeparators(dbFile));
        return false;
    }

    // Let's see if we have the correct database version
    query.prepare(QString::fromUtf8("SELECT value FROM config WHERE key = ?"));
    query.bindValue(0, DB_DB_VERSION);
    query.exec();
    query.next();
    int dbVersion = query.value(0).toInt();
    if (dbVersion != DB_VERSION) {
        m_openError = tr("Die Datenbank hat die falsche Version (ist %1, sollte %2 sein). "
                         "Entweder wurde die Datenbank mit einer zu alten Version von Muckturnier "
                         "(vor 0.5.1) erstellt, oder die Datei ist keine Muckturnier-"
                         "Datenbank.").arg(QString::number(dbVersion), QString::number(DB_VERSION));
        return false;
    }

    // Everything seems to be fine :-)
    updateSettings();
    checkIfTournamentStarted();

    return true;
}

QString Database::openError() const
{
    return m_openError;
}

QString Database::playersString(Numerus numerus, bool article) const
{
    // Returns a correct number player(s)/pair(s) string.
    QString playerString;
    if (numerus == Numerus::singular) {
        if (m_settingsTournamentMode == TournamentMode::fixedPairs) {
            if (article) {
                playerString = tr("das Paar");
            } else {
                playerString = tr("Paar");
            }
        } else {
            if (article) {
                playerString = tr("der Spieler");
            } else {
                playerString = tr("Spieler");
            }
        }
    } else if (numerus == Numerus::plural) {
        if (m_settingsTournamentMode == TournamentMode::fixedPairs) {
            if (article) {
                playerString = tr("die Paare");
            } else {
                playerString = tr("Paare");
            }
        } else {
            if (article) {
                playerString = tr("die Spieler");
            } else {
                playerString = tr("Spieler");
            }
        }
    }
    return playerString;
}

void Database::reset()
{
    m_db.transaction();
    QSqlQuery query(m_db);
    query.exec(QString::fromUtf8("DELETE FROM config"));
    query.exec(QString::fromUtf8("DELETE FROM players"));
    query.exec(QString::fromUtf8("DELETE FROM assignment"));
    query.exec(QString::fromUtf8("DELETE FROM score"));
    m_db.commit();

    setStandardValues();
    updateSettings();
    m_tournamentStarted = false;
}

void Database::saveSettings()
{
    m_db.transaction();
    QSqlQuery query(m_db);
    query.prepare(QString::fromUtf8("UPDATE config SET value = ? WHERE key = ?"));

    QVariantList values;
    switch (m_settingsTournamentMode) {
    case TournamentMode::fixedPairs:
        values << DB_FIXED_PAIRS;
        break;
    case TournamentMode::singlePlayers:
        values << DB_SINGLE_PLAYERS;
        break;
    }
    values << m_settingsBoogerScore
           << m_settingsGamesPerRound;
    query.addBindValue(values);

    QVariantList keys;
    keys << DB_TOURNAMENT_MODE
         << DB_BOOGER_SCORE
         << DB_GAMES_PER_ROUND;
    query.addBindValue(keys);

    query.execBatch();
    m_db.commit();
}

void Database::checkIfTournamentStarted()
{
    QSqlQuery query(m_db);
    query.exec(QString::fromUtf8("SELECT COUNT(round) FROM SCORE"));
    query.next();
    m_tournamentStarted = query.value(0).toInt() != 0;
}

int Database::necessaryDivisor()
{
    return 2 + (m_settingsTournamentMode != TournamentMode::fixedPairs) * 2;
}

QStringList Database::listPlayers()
{
    // Returns an alphabetically sorted list of all player/pair names.

    QStringList playersList;
    QSqlQuery query(m_db);
    query.exec(QString::fromUtf8("SELECT name FROM players ORDER BY name COLLATE NOCASE"));
    while (query.next()) {
        playersList << query.value(0).toString();
    }

    return playersList;
}

bool Database::checkName(QString name)
{
    // Checks if a player/pair name already exists.
    name = name.toLower();
    for (const QString &player : listPlayers()) {
        if (player.toLower() == name) {
            return false;
        }
    }
    return true;
}

void Database::registerPlayers(QString name)
{
    m_db.transaction();
    QSqlQuery query(m_db);
    query.prepare(QString::fromUtf8("INSERT INTO players(id, name) VALUES(NULL, ?)"));
    query.bindValue(0, name);
    query.exec();
    m_db.commit();
}

void Database::deletePlayers(QString name)
{
    m_db.transaction();
    QSqlQuery query(m_db);
    query.prepare(QString::fromUtf8("DELETE FROM players WHERE name = ?"));
    query.bindValue(0, name);
    query.exec();
    m_db.commit();
}

void Database::renamePlayers(QString oldName, QString newName)
{
    m_db.transaction();
    QSqlQuery query(m_db);
    query.prepare(QString::fromUtf8("UPDATE players SET name = ? WHERE name = ?"));
    query.bindValue(0, newName);
    query.bindValue(1, oldName);
    query.exec();
    m_db.commit();
}

int Database::playersCount()
{
    QSqlQuery query(m_db);
    query.exec(QString::fromUtf8("SELECT COUNT(id) AS count FROM players"));
    query.next();
    return query.value(0).toInt();
}


QMap<int, bool> Database::roundsState()
{
    QMap<int, bool> roundsStateMap;
    int allGames = playersCount() * m_settingsGamesPerRound;
    int roundNumber = 1;

    QSqlQuery query(m_db);
    query.exec(QString::fromUtf8("SELECT round, COUNT(round) AS count FROM score "
                                 "GROUP BY round ORDER BY round"));

    // Check if each round already contains all games
    while (query.next()) {
        roundNumber = query.value(0).toInt();
        if (query.value(1).toInt() < allGames) {
            roundsStateMap[roundNumber] = false;
        } else {
            roundsStateMap[roundNumber] = true;
        }
    }

    // Be sure to have a state for each round. All data from a round could have been deleted.
    // If we have no data, round 1 will be added as open.
    for (int i = 1; i <= roundNumber; i++) {
        if (! roundsStateMap.contains(i)) {
            roundsStateMap[i] = false;
        }
    }

    return roundsStateMap;
}

QList<int> Database::missingTables(int round)
{
    QList<int> missingTableNumbers;

    // First, we assume we need all tables
    for (int i = 1; i <= playersCount() / necessaryDivisor(); i++) {
        missingTableNumbers << i;
    }

    // Then, we take away those that already have a score
    QSqlQuery query(m_db);
    query.prepare(QString::fromUtf8("SELECT table_number FROM score WHERE round = ? "
                                    "GROUP BY table_number ORDER BY table_number"));
    query.bindValue(0, round);
    query.exec();
    while (query.next()) {
        missingTableNumbers.removeOne(query.value(0).toInt());
    }

    return missingTableNumbers;
}

QMap<QString, int> Database::missingPlayers(int round)
{
    QMap<QString, int> missingPlayersMap;
    QSqlQuery query(m_db);

    // Get a list of all players/pairs
    query.exec(QString::fromUtf8("SELECT name, id FROM players"));
    while (query.next()) {
        missingPlayersMap[query.value(0).toString()] = query.value(1).toInt();
    }

    // Get a list of all players/pairs with a score in this round and remove them
    query.prepare(QString::fromUtf8("SELECT pair_or_player FROM score "
                                    "WHERE round = ? GROUP BY pair_or_player"));
    query.bindValue(0, round);
    query.exec();
    while (query.next()) {
        missingPlayersMap.remove(missingPlayersMap.key(query.value(0).toInt()));
    }

    return missingPlayersMap;
}

Database::Assignment Database::getAssignment(int roundNumber)
{
    int round = roundNumber - 1;
    int allTables = playersCount() / necessaryDivisor();
    QSqlQuery query(m_db);
    query.prepare(QString::fromUtf8("SELECT pair_or_player, pair_number, table_number "
                                    "FROM assignment WHERE round = ?"));
    query.bindValue(0, round);
    query.exec();

    QMap<int, int> assignedTable;
    QMap<int, int> assignedPairNumber;
    QMap<int, QMap<int, int>> assignedPair;

    while(query.next()) {
        if (query.value(1).toInt() == 1) {
            assignedTable[query.value(0).toInt()] = query.value(2).toInt();
        } else {
            assignedTable[query.value(0).toInt()] = query.value(2).toInt() % allTables + 1;
        }
        assignedPairNumber[query.value(0).toInt()] = query.value(1).toInt();
    }

    for (const int &pairNumber : assignedTable.keys()) {
        assignedPair[assignedTable[pairNumber]][assignedPairNumber[pairNumber]] = pairNumber;
    }

    Assignment assignment;
    assignment.assignedTable = assignedTable;
    assignment.assignedPairNumber = assignedPairNumber;
    assignment.assignedPair = assignedPair;

    return assignment;
}

QList<Database::Score> Database::getScores(int round)
{
    QList<Score> scores;
    QSqlQuery query(m_db);
    query.prepare(QString::fromUtf8("SELECT "
                                        "score.table_number, "
                                        "players.name, "
                                        "score.score "
                                    "FROM "
                                        "score, "
                                        "players, "
                                        "assignment "
                                    "WHERE "
                                        "score.round = ? "
                                        "AND score.pair_or_player = players.id "
                                        "AND score.pair_or_player = assignment.pair_or_player "
                                        "AND score.round = assignment.round "
                                    "ORDER BY "
                                        "score.table_number, "
                                        "assignment.pair_number, "
                                        "score.game"));
    query.bindValue(0, round);
    query.exec();

    while (query.next()) {
        Score score;
        score.table = query.value(0).toInt();
        QString pairName = query.value(1).toString();
        if (m_settingsTournamentMode == TournamentMode::singlePlayers) {
            query.next();
            pairName = pairName + QString::fromUtf8(" / ") + query.value(1).toString();
        }
        score.pairName = pairName;
        score.score = query.value(2).toInt();
        scores << score;
    }

    return scores;
}

QList<Database::SingleScore> Database::getSingleScore(int round, int table)
{
    QList<SingleScore> score;
    QSqlQuery query(m_db);
    query.prepare(QString::fromUtf8("SELECT "
                                        "players.name, "
                                        "score.score "
                                    "FROM "
                                        "score, "
                                        "players, "
                                        "assignment "
                                    "WHERE "
                                        "score.round = ? "
                                        "AND score.table_number = ? "
                                        "AND score.pair_or_player = players.id "
                                        "AND score.pair_or_player = assignment.pair_or_player "
                                        "AND score.round = assignment.round "
                                    "ORDER BY "
                                        "assignment.pair_number, "
                                        "score.game"));
    query.bindValue(0, round);
    query.bindValue(1, table);
    query.exec();

    while (query.next()) {
        SingleScore singleScore;
        singleScore.pairOrPlayer = query.value(0).toString();
        singleScore.score = query.value(1).toInt();
        score << singleScore;
    }

    return score;
}

void Database::insertScore(QSqlQuery &query,
                           int round, int table, int game, int pairOrPlayer, int score)
{
    query.prepare(QString::fromUtf8("INSERT INTO score(round, table_number, game, "
                                                      "pair_or_player, score) "
                                    "VALUES(?, ?, ?, ?, ?)"));
    query.bindValue(0, round);
    query.bindValue(1, table);
    query.bindValue(2, game);
    query.bindValue(3, pairOrPlayer);
    query.bindValue(4, score);
    query.exec();
}

void Database::updateAssignment(QSqlQuery &query,
                                int round, int pairOrPlayer, int pairNumber, int table)
{
    query.prepare(QString::fromUtf8("INSERT INTO assignment(round, pair_or_player, "
                                                           "pair_number, table_number) "
                                    "VALUES(?, ?, ?, ?)"));
    query.bindValue(0, round);
    query.bindValue(1, pairOrPlayer);
    query.bindValue(2, pairNumber);
    query.bindValue(3, table);
    query.exec();
}

void Database::saveScore(int round, int table,
                         QList<int> players, QList<int> pair1Scores, QList<int> pair2Scores)
{
    m_db.transaction();
    QSqlQuery query(m_db);

    for (int game = 1; game <= pair1Scores.size(); game++) {
        insertScore(query, round, table, game, players.at(0), pair1Scores.at(game - 1));
        insertScore(query, round, table, game, players.at(1), pair2Scores.at(game - 1));
        if (m_settingsTournamentMode == TournamentMode::singlePlayers) {
            insertScore(query, round, table, game, players.at(2), pair1Scores.at(game - 1));
            insertScore(query, round, table, game, players.at(3), pair2Scores.at(game - 1));
        }

        if (game == 1) {
            updateAssignment(query, round, players.at(0), 1, table);
            updateAssignment(query, round, players.at(1), 2, table);
            if (m_settingsTournamentMode == TournamentMode::singlePlayers) {
                updateAssignment(query, round, players.at(2), 1, table);
                updateAssignment(query, round, players.at(3), 2, table);
            }
        }
    }

    m_db.commit();

    if (! m_tournamentStarted) {
        m_tournamentStarted = true;
    }
}

void Database::deleteScore(int round, int table)
{
    m_db.transaction();
    QSqlQuery query(m_db);

    query.prepare(QString::fromUtf8("DELETE FROM score WHERE round = ? AND table_number = ?"));
    query.bindValue(0, round);
    query.bindValue(1, table);
    query.exec();

    query.prepare(QString::fromUtf8("DELETE FROM assignment WHERE round = ? AND table_number = ?"));
    query.bindValue(0, round);
    query.bindValue(1, table);
    query.exec();

    m_db.commit();
    checkIfTournamentStarted();
}

QList<Database::Ranking> Database::getRanking()
{
    QSqlQuery query(m_db);
    query.prepare(QString::fromUtf8(
        "SELECT "
            "players.name AS players_name, "
            "SUM(CASE WHEN score.score = ? THEN 1 ELSE 0 END) AS boogers, "
            "SUM(CASE WHEN score.score = ? THEN 0 ELSE score.score END) AS goals, "
            "COUNT(score.score) "
        "FROM "
            "players, "
            "score "
        "WHERE "
            "players.id = score.pair_or_player "
        "GROUP BY "
            "score.pair_or_player "
        "ORDER BY "
            "boogers DESC, "
            "goals DESC, "
            "players_name"));
    query.bindValue(0, m_settingsBoogerScore);
    query.bindValue(1, m_settingsBoogerScore);
    query.exec();

    QList<Ranking> allRankings;
    QStringList namesWithScore;
    while (query.next()) {
        Ranking ranking;
        ranking.name = query.value(0).toString();
        ranking.boogers = query.value(1).toInt();
        ranking.goals = query.value(2).toInt();
        ranking.playedGames = query.value(3).toInt();
        allRankings << ranking;
        namesWithScore << ranking.name;
    }

    // Be sure to have data for all players/pairs, in case some of them don't have a score yet
    for (const QString &name : listPlayers()) {
        if (! namesWithScore.contains(name)) {
            Ranking ranking;
            ranking.name = name;
            ranking.boogers = 0;
            ranking.goals = 0;
            ranking.playedGames = 0;
            allRankings << ranking;
        }
    }

    return allRankings;
}

QString Database::getMuckturnierVersion() const
{
    return m_dbValueMuckturnierVersion;
}

Database::TournamentMode Database::getTournamentMode() const
{
    return m_settingsTournamentMode;
}

QString Database::getDbFile() const
{
    return m_db.databaseName();
}

void Database::setTournamentMode(Database::TournamentMode tournamentMode)
{
    m_settingsTournamentMode = tournamentMode;
}

int Database::getDbVersion() const
{
    return m_dbValueDbVersion;
}

int Database::getBoogerScore() const
{
    return m_settingsBoogerScore;
}

void Database::setBoogerScore(int boogerScore)
{
    m_settingsBoogerScore = boogerScore;
}

int Database::getGamesPerRound() const
{
    return m_settingsGamesPerRound;
}

void Database::setGamesPerRound(int gamesPerRound)
{
    m_settingsGamesPerRound = gamesPerRound;
}

QString Database::boogify(int score)
{
    if (score == m_settingsBoogerScore) {
        return QString::number(score);
    } else {
        return QString::fromUtf8("%1 ●").arg(QString::number(score));
    }
}
