/* Copyright (c) 2012-2017 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef DATABASE_H
#define DATABASE_H

// Qt includes
#include <QObject>
#include <QSqlDatabase>
#include <QMap>

// Qt classes
class QSqlQuery;

class Database : public QObject
{

public:
    enum TournamentMode {
        fixedPairs,
        singlePlayers
    };

    enum Numerus {
        singular,
        plural
    };

    struct Assignment {
        QMap<int, int> assignedTable;
        QMap<int, int> assignedPairNumber;
        QMap<int, QMap<int, int>> assignedPair;
    };

    struct Score {
        int table;
        QString pairName;
        int score;
    };

    struct SingleScore {
        QString pairOrPlayer;
        int score;
    };

    struct Ranking {
        QString name;
        int boogers;
        int goals;
        int playedGames;
    };

    explicit Database(QObject *parent);
    ~Database();
    bool create(QString dbFile);
    bool tournamentStarted();
    bool openDb(QString dbFile);
    QString openError() const;
    QString playersString(Numerus numerus, bool article = false) const;
    void reset();
    void saveSettings();
    void deletePlayers(QString name);
    QStringList listPlayers();
    int necessaryDivisor();
    bool checkName(QString name);
    void registerPlayers(QString name);
    void renamePlayers(QString oldName, QString newName);
    QMap<int, bool> roundsState();
    QList<int> missingTables(int round);
    QMap<QString, int> missingPlayers(int round);
    Assignment getAssignment(int roundNumber);
    QList<Score> getScores(int round);
    QList<SingleScore> getSingleScore(int round, int table);
    QString boogify(int score);
    void saveScore(int round, int table, QList<int> players,
                   QList<int> pair1Scores, QList<int> pair2Scores);
    void deleteScore(int round, int table);
    QList<Ranking> getRanking();
    QString getMuckturnierVersion() const;
    int getDbVersion() const;
    QString getDbFile() const;
    Database::TournamentMode getTournamentMode() const;
    int getBoogerScore() const;
    int getGamesPerRound() const;
    int playersCount();
    void setTournamentMode(Database::TournamentMode tournamentMode);
    void setBoogerScore(int boogerScore);
    void setGamesPerRound(int gamesPerRound);
    void closeDatabaseConnection();

private: // Functions
    bool connectDatabase(QString &dbFile);
    void updateSettings();
    void setStandardValues();
    void checkIfTournamentStarted();
    void insertScore(QSqlQuery &query, int round, int table, int game, int pairOrPlayer, int score);
    void updateAssignment(QSqlQuery &query, int round, int pairOrPlayer, int pairNumber, int table);

private: // Variables
    QSqlDatabase m_db;
    bool m_tournamentStarted;
    QString m_openError;
    QString m_dbValueMuckturnierVersion;
    int m_dbValueDbVersion;
    TournamentMode m_settingsTournamentMode;
    int m_settingsBoogerScore;
    int m_settingsGamesPerRound;

};

#endif // DATABASE_H
