/* Copyright (c) 2012-2017 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "ExportDialog.h"
#include "Database.h"
#include "RankingPage.h"
#include "version.h"

// Qt includes
#include <QDebug>
#include <QGroupBox>
#include <QLabel>
#include <QHBoxLayout>
#include <QVBoxLayout>
#include <QPushButton>
#include <QCheckBox>
#include <QLineEdit>
#include <QFileDialog>
#include <QDir>
#include <QStandardPaths>
#include <QMessageBox>
#include <QApplication>
#include <QTextCodec>
#include <QDateTime>
#include <QDesktopServices>
#include <QTableWidget>
#include <QRadioButton>
#include <QRegularExpression>
#include <QDialogButtonBox>

ExportDialog::ExportDialog(QWidget *parent, Database *db, RankingPage *rankingPage)
    : QDialog(parent), m_db(db), m_rankingPage(rankingPage)
{
    setAttribute(Qt::WA_DeleteOnClose);
    setWindowTitle(tr("Muckturnier: Turnierdaten exportieren"));

    QVBoxLayout *layout = new QVBoxLayout(this);
    layout->setSizeConstraint(QLayout::SetFixedSize);

    QGroupBox *optionsBox = new QGroupBox(tr("Exportoptionen"));
    QVBoxLayout *optionsBoxLayout = new QVBoxLayout(optionsBox);
    layout->addWidget(optionsBox);

    // Title

    QHBoxLayout *titleEntryLayout = new QHBoxLayout;
    optionsBoxLayout->addLayout(titleEntryLayout);
    m_exportTitle = new QCheckBox(tr("Turniertitel ausgeben:"));
    m_exportTitle->setChecked(Qt::Checked);
    titleEntryLayout->addWidget(m_exportTitle);

    m_titleEdit = new QLineEdit;
    titleEntryLayout->addWidget(m_titleEdit);
    connect(m_titleEdit, &QLineEdit::textEdited, this, &ExportDialog::checkTitle);

    // Show options or not

    QCheckBox *customExport = new QCheckBox(tr("Export anpassen"));
    customExport->setChecked(Qt::Unchecked);
    optionsBoxLayout->addWidget(customExport);
    connect(customExport, &QCheckBox::stateChanged, this, &ExportDialog::setExportType);

    // Formatting

    m_formatBox = new QGroupBox(tr("Formatierung"));
    QVBoxLayout *formatBoxLayout = new QVBoxLayout(m_formatBox);
    m_formatBox->hide();
    optionsBoxLayout->addWidget(m_formatBox);

    m_exportHeadlines = new QCheckBox(tr("Überschriften der Datensätze anzeigen "
                                         "(z.\u2009B. „Einzelergebnisse“)"));
    m_exportHeadlines->setChecked(Qt::Checked);
    formatBoxLayout->addWidget(m_exportHeadlines);

    m_exportSubHeadlines = new QCheckBox(tr("Unterüberschriften anzeigen (z.\u2009B. "
                                            "„Spielergebnisse der 1. Runde“)"));
    m_exportSubHeadlines->setChecked(Qt::Checked);
    formatBoxLayout->addWidget(m_exportSubHeadlines);

    m_displayBoogers = new QCheckBox(tr("Bobbl-Symbole anzeigen"));
    m_displayBoogers->setChecked(Qt::Checked);
    formatBoxLayout->addWidget(m_displayBoogers);

    // Datasets to export

    m_selectDataBox = new QGroupBox(tr("Zu exportierende Datensätze"));
    QVBoxLayout *selectDataBoxLayout = new QVBoxLayout(m_selectDataBox);
    m_selectDataBox->hide();
    optionsBoxLayout->addWidget(m_selectDataBox);

    m_exportGenericData = new QCheckBox(tr("Allgemeine Daten (Datum, Anzahl Teilnehmer etc.)"));
    m_exportGenericData->setChecked(Qt::Checked);
    selectDataBoxLayout->addWidget(m_exportGenericData);
    connect(m_exportGenericData, &QCheckBox::stateChanged,
            this, &ExportDialog::checkExportPossible);

    m_exportRanking = new QCheckBox(tr("Rangliste"));
    m_exportRanking->setChecked(Qt::Checked);
    selectDataBoxLayout->addWidget(m_exportRanking);
    connect(m_exportRanking, &QCheckBox::stateChanged, this, &ExportDialog::checkExportPossible);

    m_exportRoundResults = new QCheckBox;
    m_exportRoundResults->setChecked(Qt::Checked);
    selectDataBoxLayout->addWidget(m_exportRoundResults);
    connect(m_exportRoundResults, &QCheckBox::stateChanged,
            this, &ExportDialog::checkExportPossible);

    QHBoxLayout *roundSelectLayout = new QHBoxLayout;
    selectDataBoxLayout->addLayout(roundSelectLayout);

    m_roundSelect = new QLineEdit;
    m_defaultTextColor = m_roundSelect->palette().color(QPalette::Foreground);
    roundSelectLayout->addWidget(m_roundSelect);
    connect(m_roundSelect, &QLineEdit::textChanged, this, &ExportDialog::checkRoundSelection);

    m_roundSelectLabel = new QLabel(tr("Format z.\u2009B.: „1, 3, 7-9, 11“"));
    roundSelectLayout->addWidget(m_roundSelectLabel);

    m_errorLabel = new QLabel;
    selectDataBoxLayout->addWidget(m_errorLabel);
    m_errorLabel->hide();

    layout->addStretch();

    QGroupBox *fileFormatBox = new QGroupBox(tr("Dateiformat"));
    QVBoxLayout *fileFormatBoxLayout = new QVBoxLayout(fileFormatBox);
    layout->addWidget(fileFormatBox);

    m_exportHTML = new QRadioButton(tr("HTML"));
    m_exportHTML->setChecked(Qt::Checked);
    fileFormatBoxLayout->addWidget(m_exportHTML);

    m_exportCSV = new QRadioButton(tr("CSV (UTF-8 mit Tabulatoren als Trenner)"));
    fileFormatBoxLayout->addWidget(m_exportCSV);

    QGroupBox *actionsBox = new QGroupBox(tr("Aktionen"));
    QVBoxLayout *actionsBoxLayout = new QVBoxLayout(actionsBox);
    layout->addWidget(actionsBox);

    m_openAfterExport = new QCheckBox(tr("Datei nach dem Export öffnen"));
    m_openAfterExport->setChecked(Qt::Checked);
    actionsBoxLayout->addWidget(m_openAfterExport);

    m_closeAfterExport = new QCheckBox(tr("Dialog nach dem Export schließen"));
    m_closeAfterExport->setChecked(Qt::Checked);
    actionsBoxLayout->addWidget(m_closeAfterExport);

    // The dialog's buttons
    m_buttonBox = new QDialogButtonBox(QDialogButtonBox::Ok | QDialogButtonBox::Cancel);
    m_buttonBox->button(QDialogButtonBox::Ok)->setText(tr("Exportieren"));
    m_buttonBox->button(QDialogButtonBox::Cancel)->setText(tr("Schließen"));
    layout->addWidget(m_buttonBox);
    connect(m_buttonBox, &QDialogButtonBox::accepted, this, &QDialog::accept);
    connect(m_buttonBox, &QDialogButtonBox::rejected, this, &QDialog::reject);

    // Fill in data

    m_dateFormat = QLocale().dateFormat(QLocale::ShortFormat);
    if (! m_dateFormat.contains(QString::fromUtf8("yyyy"))) {
        m_dateFormat = m_dateFormat.replace(QString::fromUtf8("yy"), QString::fromUtf8("yyyy"));
    }
    m_timeFormat = QLocale().timeFormat(QLocale::ShortFormat);

    if (m_rankingPage->m_allRounds == 1) {
        m_exportRoundResults->setText(tr("Einzelergebnisse der folgenden Runden "
                                         "(Es wurde eine Runde gespielt)"));
        m_roundSelect->setText(QString::fromUtf8("1"));
    } else {
        QString lastRound = QString::number(m_rankingPage->m_allRounds);
        m_exportRoundResults->setText(tr("Einzelergebnisse der folgenden Runden "
                                         "(Es wurden %1 Runden gespielt)").arg(lastRound));
        m_roundSelect->setText(QString::fromUtf8("1-") + lastRound);
    }

    // Add the database's filename as the default title and set a reasonable width
    m_titleEdit->setText(QFileInfo(m_db->getDbFile()).baseName());
    m_titleEdit->setMinimumSize(m_titleEdit->sizeHint().width() * 2,
                                m_titleEdit->sizeHint().height());

    m_titleEdit->setFocus();
    m_titleEdit->selectAll();
}

void ExportDialog::setExportType(bool custom)
{
    if (custom) {
        m_selectDataBox->show();
        m_formatBox->show();
    } else {
        m_exportTitle->setChecked(Qt::Checked);
        m_exportHeadlines->setChecked(Qt::Checked);
        m_exportSubHeadlines->setChecked(Qt::Checked);
        m_displayBoogers->setChecked(Qt::Checked);
        m_exportGenericData->setChecked(Qt::Checked);
        m_exportRanking->setChecked(Qt::Checked);
        m_exportRoundResults->setChecked(Qt::Checked);

        if (m_rankingPage->m_allRounds == 1) {
            m_roundSelect->setText(QString::fromUtf8("1"));
        } else {
            m_roundSelect->setText(QString::fromUtf8("1-")
                                   + QString::number(m_rankingPage->m_allRounds));
        }
        m_selectDataBox->hide();
        m_formatBox->hide();
    }
}

void ExportDialog::checkTitle(QString title)
{
    m_exportTitle->setChecked(! title.simplified().isEmpty());
    m_exportTitle->setEnabled(! title.simplified().isEmpty());
}

void ExportDialog::roundSelectionIsInvalid(QString message)
{
    if (! message.isEmpty()) {
        m_errorLabel->show();
        m_errorLabel->setText(message);
    }
    setRoundSelectColor(Qt::red);
    m_roundSelectionIsRed = true;
    m_exportRoundResults->setEnabled(false);
    m_exportRoundResults->setChecked(false);
    checkExportPossible();
}

void ExportDialog::checkRoundSelection(QString range)
{
    m_selectedRounds.clear();

    // No input: range is not okay, but set the default color
    if (range.isEmpty()) {
        roundSelectionIsInvalid();
        setRoundSelectColor(m_defaultTextColor);
        m_errorLabel->hide();
        return;
    }

    // Remove all whitespace
    QString parsedRange = range.simplified().remove(QString::fromUtf8(" "));

    // Search for input other than 0-9, "," and "-"
    QRegularExpression check(QString::fromUtf8("([^\\d,-]+)"));
    QRegularExpressionMatch match = check.match(parsedRange);
    if (match.hasMatch()) {
        roundSelectionIsInvalid(tr("Ungültige Eingabe: „%1“").arg(match.captured()));
        return;
    }

    // Process all parts
    for (const QString &part : parsedRange.split(QString::fromUtf8(","))) {

        // Single round
        if (! part.contains(QString::fromUtf8("-"))) {
            // Skip empty parts
            if (part.isEmpty()) {
                continue;
            }

            // Check the requested round
            int requestedRound = part.toInt();
            if (requestedRound > 0 && requestedRound <= m_rankingPage->m_allRounds) {
                if (! m_selectedRounds.contains(requestedRound)) {
                    m_selectedRounds << requestedRound;
                }
            } else {
                roundSelectionIsInvalid(tr("Ungültige Rundenzahl: %1").arg(
                                           QString::number(requestedRound)));
                return;
            }

        // Range of rounds
        } else {
            QStringList givenRange = part.split(QString::fromUtf8("-"));

            // We have to find exacly two parts
            if (givenRange.size() != 2) {
                roundSelectionIsInvalid(tr("Ungültige Intervallangabe: „%1“").arg(part));
                return;
            }

            // About to enter, e. g. "1-", handle like a single round entry
            if (givenRange[1].isEmpty()) {
                int requestedRound = givenRange[0].toInt();
                if (requestedRound > 0 && requestedRound <= m_rankingPage->m_allRounds) {
                    if (! m_selectedRounds.contains(requestedRound)) {
                        m_selectedRounds << requestedRound;
                    }
                } else {
                    roundSelectionIsInvalid(tr("Ungültige Rundenzahl: „%1“").arg(
                                               QString::number(requestedRound)));
                    return;
                }

            // True range entered
            } else {
                int fromRound = givenRange[0].toInt();
                int toRound = givenRange[1].toInt();

                // Check the range
                if (fromRound == 0 || fromRound > m_rankingPage->m_allRounds
                    || toRound == 0 || toRound > m_rankingPage->m_allRounds
                    || fromRound > toRound) {
                    roundSelectionIsInvalid(tr("Ungültiges Intervall: „%1“").arg(part));
                    return;
                }

                // Add all rounds between
                for (int requestedRound = fromRound; requestedRound <= toRound; requestedRound++) {
                    if (! m_selectedRounds.contains(requestedRound)) {
                        m_selectedRounds << requestedRound;
                    }
                }
            }
        }
    }

    if (m_roundSelectionIsRed) {
        setRoundSelectColor(m_defaultTextColor);
        m_roundSelectionIsRed = false;
        m_errorLabel->hide();
    }

    m_exportRoundResults->setEnabled(true);
    m_exportRoundResults->setChecked(true);

    qSort(m_selectedRounds);
    checkExportPossible();
}

void ExportDialog::checkExportPossible()
{
    m_buttonBox->button(QDialogButtonBox::Ok)->setEnabled(m_exportGenericData->isChecked()
                                                          || m_exportRanking->isChecked()
                                                          || m_exportRoundResults->isChecked());
}

void ExportDialog::setRoundSelectColor(QColor color)
{
    QPalette palette;
    palette.setColor(QPalette::Text, color);
    m_roundSelect->setPalette(palette);
}

void ExportDialog::accept()
{
    QString type;
    QString extension;
    QString message;
    if (m_exportHTML->isChecked()) {
        type = tr("HTML-Dateien");
        extension = QString::fromUtf8("htm");
        message = tr("Exportiere HTML-Daten …");
    } else {
        type = tr("CSV-Dateien");
        extension = QString::fromUtf8("csv");
        message = tr("Exportiere CSV-Daten …");
    }

    QString title = m_titleEdit->text().simplified();
    if (title.isEmpty()) {
        title = tr("Protokoll");
    }

    QString journalFileName = QFileDialog::getSaveFileName(this,
        tr("Turnierdaten-Export: Bitte einen Dateinamen auswählen"),
        QDir::toNativeSeparators(QStandardPaths::writableLocation(QStandardPaths::DocumentsLocation)
                                 + QString::fromUtf8("/") + title
                                 + QString::fromUtf8(".%1").arg(extension)),
        type + QString::fromUtf8(" (*.%1);; Alle Dateien (*)").arg(extension));

    if (journalFileName.isEmpty()) {
        return;
    }

    QFile journalFile(journalFileName);
    if (! journalFile.open(QIODevice::WriteOnly | QIODevice::Text)) {
        QMessageBox::critical(this,
            tr("Turnierdaten-Export"),
            tr("Es bestehen keine Schreibrechte für die Datei „%1“. "
               "Bitte die Zugriffsrechte überprüfen!").arg(
               QDir::toNativeSeparators(journalFileName)),
            QMessageBox::Ok);
        return;
    }

    QTextStream outStream(&journalFile);
    outStream.setCodec(QTextCodec::codecForName("UTF-8"));
    emit statusUpdate(message);
    QApplication::setOverrideCursor(Qt::WaitCursor);

    if (m_exportHTML->isChecked()) {
        saveHTMLFile(outStream);
    } else {
        saveCSVFile(outStream);
    }

    journalFile.close();

    if (m_openAfterExport->isChecked()) {
        QDesktopServices::openUrl(QUrl::fromLocalFile(journalFileName));
    }

    QApplication::restoreOverrideCursor();
    emit statusUpdate(tr("Daten in Datei „%1“ exportiert").arg(
                         QDir::toNativeSeparators(journalFileName)));

    if (m_closeAfterExport->isChecked()) {
        close();
    }
}

void ExportDialog::saveHTMLFile(QTextStream &outStream)
{
    // Add the HTML header
    outStream << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n\n"
              << "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.1//EN\"\n"
              << "   \"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd\">\n\n"
              << "<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\">\n\n"
              << "<head>\n\n"
              << "<title>" << m_titleEdit->text().simplified().toHtmlEscaped() << "</title>\n\n"
              << "<meta http-equiv=\"content-type\" content=\"text/html; charset=UTF-8\" />\n\n"
              << "<style type=\"text/css\">\n"
              << "table { border-spacing: 0px; }\n"
              << "td { padding: 0.2em 1em 0.2em; }\n"
              << ".line { border-top: 1px solid black; }\n"
              << "</style>\n\n"
              << "</head>\n\n"
              << "<body>\n\n";

    if (m_exportTitle->isChecked()) {
        outStream << "<h1>" << m_titleEdit->text().simplified().toHtmlEscaped() << "</h1>\n\n";
    }

    if (m_exportGenericData->isChecked()) {
        // Add tournament meta data

        if (m_exportHeadlines->isChecked()) {
            outStream << "<h2>" << tr("Allgemeine Daten") << "</h2>\n\n";
        }
        outStream << "<p>" << tr("Protokoll erstellt am %1 um %2 Uhr.").arg(
                                 QDateTime::currentDateTime().toString(m_dateFormat),
                                 QDateTime::currentDateTime().toString(m_timeFormat)) << "</p>\n\n";

        if (m_exportSubHeadlines->isChecked()) {
            outStream << "<h3>Muckturnier</h3>\n\n";
        }

        outStream << "<table>\n";
        if (QString::fromUtf8(MT_VERSION) == m_db->getMuckturnierVersion()) {
            outStream << "<tr>"
                      << "<td>" << tr("Turnierdatenbank und Protokoll erstellt mit:") << "</td>"
                      << "<td>Muckturnier " << MT_VERSION << "</td>"
                      << "</tr>\n";
        } else {
            outStream << "<tr><td>" << tr("Turnierdatenbank erstellt mit:")
                      << "</td><td>Muckturnier " << m_db->getMuckturnierVersion() << "</td></tr>\n"
                      << "<tr><td>" << tr("Protokoll erstellt mit:")
                      << "</td><td>Muckturnier "
                      << MT_VERSION << "</td></tr>\n";
        }
        outStream << "<tr><td>" << tr("Homepage:") << "</td>"
                  << "<td><a href=\"http://nasauber.de/opensource/muckturnier/\">"
                  << "http://nasauber.de/opensource/muckturnier/</a></td></tr>\n"
                  << "</table>\n\n";

        if (m_exportSubHeadlines->isChecked()) {
            outStream << "<h3>" << tr("Turnier") << "</h3>\n\n";
        }

        outStream << "<table>\n"
                  << "<tr><td>" << tr("Turniertyp:") << "</td><td>";
        switch (m_db->getTournamentMode()) {
        case Database::TournamentMode::fixedPairs:
            outStream << tr("Feste Paare");
            break;
        case Database::TournamentMode::singlePlayers:
            outStream << tr("Einzelne Spieler");
            break;
        }
        outStream << "</td></tr>\n"
                  << "<tr><td>" << tr("Teilnehmende %1:").arg(
                                      m_db->playersString(Database::Numerus::plural)) << "</td><td>"
                  << QString::number(m_rankingPage->m_ranking->rowCount()) << "</td></tr>\n"
                  << "<tr><td>" << tr("Gespielte Runden:") << "</td><td>"
                  << QString::number(m_rankingPage->m_allRounds) << "</td></tr>\n"
                  << "</table>\n\n";
    }

    if (m_exportRanking->isChecked()) {
        // Add the ranking

        if (m_exportHeadlines->isChecked()) {
            outStream << "<h2>" << m_rankingPage->m_rankingBox->title() << "</h2>\n\n";
        }
        outStream << "<table>\n"
                  << "<thead>\n"
                  << "<tr><td><b>" << tr("Platz") << "</b></td>"
                  << "<td><b>" << m_db->playersString(Database::Numerus::singular) << "</b></td>"
                  << "<td><b>" << tr("Bobbl") << "</b>"
                  << "</td><td><b>" << tr("Tore") << "</b></td></tr>\n"
                  << "</thead>\n"
                  << "<tbody>\n";

        int rank;
        int prevRank = -1;
        QString style;

        for (int row = 0; row < m_rankingPage->m_ranking->rowCount(); row++) {
            rank = m_rankingPage->m_ranking->item(row, 0)->text().toInt();
            if (rank != prevRank) {
                style = QString::fromUtf8(" class=\"line\"");
                prevRank = rank;
            } else {
                style.clear();
            }

            outStream << "<tr><td" << style << ">" << QString::number(rank) << "</td>";
            for (int i = 1; i < 4; i++) {
                outStream << "<td" << style << ">"
                          << m_rankingPage->m_ranking->item(row, i)->text().toHtmlEscaped()
                          << "</td>";
            }
            outStream << "</tr>\n";
        }
        outStream << "</tbody>\n"
                  << "</table>\n\n";
    }

    if (m_exportRoundResults->isChecked()) {
        // Add all requested round data

        if (m_exportHeadlines->isChecked()) {
            outStream << "<h2>" << tr("Einzelergebnisse") << "</h2>\n\n";
        }

        for (const int &round : m_selectedRounds) {
            if (m_exportSubHeadlines->isChecked()) {
                outStream << "<h3>" << tr("Spielergebnisse der %1. Runde").arg(
                                          QString::number(round))
                          << "</h3>\n\n";
            }
            outStream << "<table>\n"
                      << "<thead>\n"
                      << "<tr>"
                      << "<td><b>" << tr("Tisch") << "</b></td>"
                      << "<td><b>" << tr("Paar") << "</b></td>";

            for (int game = 1; game <= m_db->getGamesPerRound(); game++) {
                outStream << "<td><b>" << tr("%1. Bobbl").arg(QString::number(game)) << "</b></td>";
            }
            outStream << "</tr>\n"
                      << "</thead>\n"
                      << "<tbody>\n";

            const QList<Database::Score> allScores = m_db->getScores(round);
            for (int i = 0; i < allScores.size(); i++) {
                QString tableNumber = QString::number(allScores[i].table);
                QString pair1Name = allScores[i].pairName;
                QStringList pair1Scores;
                for (int j = 0; j < m_db->getGamesPerRound(); j++) {
                    if (m_displayBoogers->isChecked()) {
                        pair1Scores << m_db->boogify(allScores[i].score);
                    } else {
                        pair1Scores << QString::number(allScores[i].score);
                    }
                    i++;
                }
                QString pair2Name = allScores[i].pairName;
                QStringList pair2Scores;
                for (int j = 0; j < m_db->getGamesPerRound(); j++) {
                    if (m_displayBoogers->isChecked()) {
                        pair2Scores << m_db->boogify(allScores[i].score);
                    } else {
                        pair2Scores << QString::number(allScores[i].score);
                    }
                    i++;
                }
                i--;

                outStream << "<tr><td rowspan=\"2\" class=\"line\">" << tableNumber << "</td>"
                          << "<td class=\"line\">" << pair1Name.toHtmlEscaped() << "</td>";
                for (const QString &score : pair1Scores) {
                    outStream << "<td class=\"line\">" << score << "</td>";
                }
                outStream << "</tr>\n<tr><td>" << pair2Name.toHtmlEscaped() << "</td>";
                for (const QString &score : pair2Scores) {
                    outStream << "<td>" << score << "</td>";
                }
                outStream << "</tr>\n";
            }
            outStream << "</tbody>\n"
                      << "</table>\n\n";
        }
    }

    // Close HTML
    outStream << "</body>\n\n"
              << "</html>\n";
}

void ExportDialog::saveCSVFile(QTextStream &outStream)
{
    if (m_exportTitle->isChecked()) {
        outStream << m_titleEdit->text().simplified() << "\n\n";
    }

    if (m_exportGenericData->isChecked()) {
        // Add tournament meta data

        if (m_exportHeadlines->isChecked()) {
            outStream << tr("Allgemeine Daten") << "\n\n";
        }
        outStream << tr("Protokoll erstellt am %1 um %2 Uhr.").arg(
                        QDateTime::currentDateTime().toString(m_dateFormat),
                        QDateTime::currentDateTime().toString(m_timeFormat))
                  << "\n\n";

        if (m_exportSubHeadlines->isChecked()) {
            outStream << "Muckturnier\n";
        }
        if (QString::fromUtf8(MT_VERSION) == m_db->getMuckturnierVersion()) {
            outStream << tr("Turnierdatenbank und Protokoll erstellt mit:") << "\t"
                      << "Muckturnier " << MT_VERSION << "\n";
        } else {
            outStream << tr("Turnierdatenbank erstellt mit:") << "\tMuckturnier "
                      << m_db->getMuckturnierVersion() << "\n"
                      << tr("Protokoll erstellt mit:") << "\tMuckturnier "
                      << MT_VERSION << "\n";
        }
        outStream << tr("Homepage:") << "\t"
                  << "http://nasauber.de/opensource/muckturnier/\n\n";

        if (m_exportSubHeadlines->isChecked()) {
            outStream << tr("Turnier") << "\n\n";
        }

        outStream << tr("Turniertyp:") << "\t";
        switch (m_db->getTournamentMode()) {
        case Database::TournamentMode::fixedPairs:
            outStream << tr("Feste Paare");
            break;
        case Database::TournamentMode::singlePlayers:
            outStream << tr("Einzelne Spieler");
            break;
        }

        outStream << "\n"
                  << tr("Teilnehmende %1:").arg(m_db->playersString(Database::Numerus::plural))
                  << "\t" << QString::number(m_rankingPage->m_ranking->rowCount()) << "\n"
                  << tr("Gespielte Runden:") << "\t"
                  << QString::number(m_rankingPage->m_allRounds) << "\n\n";
    }

    if (m_exportRanking->isChecked()) {
        // Add the ranking

        if (m_exportHeadlines->isChecked()) {
            outStream << m_rankingPage->m_rankingBox->title() << "\n\n";
        }
        outStream << tr("Platz") << "\t"
                  << m_db->playersString(Database::Numerus::singular) << "\t"
                  << tr("Bobbl") << "\t"
                  << tr("Tore") << "\n";

        for (int row = 0; row < m_rankingPage->m_ranking->rowCount(); row++) {
            outStream << m_rankingPage->m_ranking->item(row, 0)->text();
            for (int i = 1; i < 4; i++) {
                outStream << "\t" << m_rankingPage->m_ranking->item(row, i)->text().toHtmlEscaped();
            }
            outStream << "\n";
        }
        outStream << "\n";
    }

    if (m_exportRoundResults->isChecked()) {
        // Add all requested round data

        if (m_exportHeadlines->isChecked()) {
            outStream << tr("Einzelergebnisse") << "\n\n";
        }

        for (const int &round : m_selectedRounds) {
            if (m_exportSubHeadlines->isChecked()) {
                outStream << tr("Spielergebnisse der %1. Runde").arg(QString::number(round))
                          << "\n";
            }
            outStream << tr("Tisch") << "\t" << tr("Paar");
            for (int game = 1; game <= m_db->getGamesPerRound(); game++) {
                outStream << "\t" << tr("%1. Bobbl").arg(QString::number(game));
            }
            outStream << "\n";

            const QList<Database::Score> allScores = m_db->getScores(round);
            for (int i = 0; i < allScores.size(); i++) {
                QString tableNumber = QString::number(allScores[i].table);
                QString pair1Name = allScores[i].pairName;
                QStringList pair1Scores;
                for (int j = 0; j < m_db->getGamesPerRound(); j++) {
                    if (m_displayBoogers->isChecked()) {
                        pair1Scores << m_db->boogify(allScores[i].score);
                    } else {
                        pair1Scores << QString::number(allScores[i].score);
                    }
                    i++;
                }
                QString pair2Name = allScores[i].pairName;
                QStringList pair2Scores;
                for (int j = 0; j < m_db->getGamesPerRound(); j++) {
                    if (m_displayBoogers->isChecked()) {
                        pair2Scores << m_db->boogify(allScores[i].score);
                    } else {
                        pair2Scores << QString::number(allScores[i].score);
                    }
                    i++;
                }
                i--;

                outStream << tableNumber << "\t" << pair1Name;
                for (const QString &score : pair1Scores) {
                    outStream << "\t" << score;
                }
                outStream << "\n\t" << pair2Name;
                for (const QString &score : pair2Scores) {
                    outStream << "\t" << score;
                }
                outStream << "\n";
            }
            outStream << "\n";
        }
    }
}
