/* Copyright (c) 2012-2017 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "MainWindow.h"
#include "Database.h"
#include "AboutWindow.h"
#include "SettingsDialog.h"
#include "PlayersPage.h"
#include "ScorePage.h"
#include "RankingPage.h"
#include "ExportDialog.h"
#include "StopWatch.h"

// Qt includes
#include <QLabel>
#include <QHBoxLayout>
#include <QFont>
#include <QTabWidget>
#include <QMenuBar>
#include <QAction>
#include <QApplication>
#include <QStatusBar>
#include <QMessageBox>
#include <QUrl>
#include <QDesktopServices>
#include <QFileInfo>
#include <QDebug>
#include <QFileDialog>
#include <QDir>
#include <QStandardPaths>
#include <QCoreApplication>
#include <QDirIterator>
#include <QDockWidget>
#include <QDesktopWidget>
#include <QPushButton>
#include <QVBoxLayout>

namespace {
const QString SPLASHBUTTONSTYLE = QString::fromUtf8("QPushButton { border:none; }");
}

MainWindow::MainWindow()
{
    // Setup the database
    m_db = new Database(this);

    // Window setup
    setWindowIcon(QIcon(resourcePath(QString::fromUtf8("muckturnier.svg"))));
    setWindowTitle(tr("Muckturnier"));

    QRect screenGeometry = QApplication::desktop()->screenGeometry();
    if (screenGeometry.width() >= 800 && screenGeometry.height() >= 600) {
        resize(800, 600);
    }

    m_centralWidget = new QWidget(this);
    setCentralWidget(m_centralWidget);
    QHBoxLayout *layout = new QHBoxLayout(m_centralWidget);

    // Startup buttons

    layout->addStretch();
    QVBoxLayout *newTournamentLayout = new QVBoxLayout;
    layout->addLayout(newTournamentLayout);
    newTournamentLayout->addStretch();

    QPushButton *newTournamentButton = new QPushButton;
    newTournamentButton->setStyleSheet(SPLASHBUTTONSTYLE);
    QIcon newIcon = QIcon(resourcePath(QString::fromUtf8("new.svg")));
    newTournamentButton->setIcon(newIcon);
    newTournamentButton->setIconSize(QSize(100, 100));
    connect(newTournamentButton, &QPushButton::clicked, this, &MainWindow::startNewTournament);
    newTournamentLayout->addWidget(newTournamentButton);

    QPushButton *newTournamentLabel = new QPushButton(tr("Ein neues Turnier starten"));
    connect(newTournamentLabel, &QPushButton::clicked, this, &MainWindow::startNewTournament);
    newTournamentLabel->setStyleSheet(SPLASHBUTTONSTYLE);
    newTournamentLayout->addWidget(newTournamentLabel);

    newTournamentLayout->addStretch();
    layout->addStretch();
    QVBoxLayout *loadTournamentLayout = new QVBoxLayout;
    layout->addLayout(loadTournamentLayout);
    loadTournamentLayout->addStretch();

    QPushButton *loadTournamentButton = new QPushButton;
    loadTournamentButton->setStyleSheet(SPLASHBUTTONSTYLE);
    QIcon loadIcon = QIcon(resourcePath(QString::fromUtf8("load.svg")));
    loadTournamentButton->setIcon(loadIcon);
    loadTournamentButton->setIconSize(QSize(100, 100));
    connect(loadTournamentButton, &QPushButton::clicked, this, &MainWindow::triggerLoadTournament);
    loadTournamentLayout->addWidget(loadTournamentButton);

    QPushButton *loadTournamentLabel = new QPushButton(tr("Ein Turnier öffnen"));
    connect(loadTournamentLabel, &QPushButton::clicked, this, &MainWindow::triggerLoadTournament);
    loadTournamentLabel->setStyleSheet(SPLASHBUTTONSTYLE);
    loadTournamentLayout->addWidget(loadTournamentLabel);

    loadTournamentLayout->addStretch();
    layout->addStretch();

    // Settings dialog

    m_settingsDialog = new SettingsDialog(this, m_db);
    m_settingsDialog->setWindowTitle(tr("Muckturnier: Einstellungen"));
    connect(m_settingsDialog, &SettingsDialog::statusUpdate, this, &MainWindow::updateStatus);
    connect(m_settingsDialog, &SettingsDialog::settingsSaved, this, &MainWindow::settingsSaved);
    connect(m_settingsDialog, &SettingsDialog::dbReset, this, &MainWindow::dbReset);

    // Dock setup

    m_playersPage = new PlayersPage(this, m_db);
    connect(m_playersPage, &PlayersPage::statusUpdate, this, &MainWindow::updateStatus);
    connect(m_playersPage, &PlayersPage::tournamentCanBeStarted,
            this, &MainWindow::setScoreTabEnabled);
    m_playersDock = new QDockWidget(tr("Paare"));
    m_playersDock->setWidget(m_playersPage);
    formatDockWidget(m_playersDock);
    addDockWidget(Qt::TopDockWidgetArea, m_playersDock);
    m_playersDock->hide();

    m_scorePage = new ScorePage(this, m_db);
    connect(m_scorePage, &ScorePage::tournamentStarted, this, &MainWindow::tournamentStarted);
    connect(m_scorePage, &ScorePage::statusUpdate, this, &MainWindow::updateStatus);
    m_scoreDock = new QDockWidget(tr("Ergebnisse"));
    m_scoreDock->setWidget(m_scorePage);
    formatDockWidget(m_scoreDock);
    addDockWidget(Qt::TopDockWidgetArea, m_scoreDock);
    tabifyDockWidget(m_playersDock, m_scoreDock);
    m_scoreDock->hide();

    m_rankingPage = new RankingPage(this, m_db);
    connect(m_rankingPage, &RankingPage::statusUpdate, this, &MainWindow::updateStatus);
    connect(m_playersPage, &PlayersPage::playersListChanged, m_rankingPage, &RankingPage::reload);
    connect(m_scorePage, &ScorePage::scoreChanged, m_rankingPage, &RankingPage::reload);
    m_rankingDock = new QDockWidget(tr("Rangliste"));
    m_rankingDock->setWidget(m_rankingPage);
    formatDockWidget(m_rankingDock);
    addDockWidget(Qt::TopDockWidgetArea, m_rankingDock);
    tabifyDockWidget(m_scoreDock, m_rankingDock);
    m_rankingDock->hide();

    // Menu

    QMenu *fileMenu = menuBar()->addMenu(tr("Datei"));

    QAction *newTournamentAction = new QAction(tr("Neues Turnier starten"), this);
    fileMenu->addAction(newTournamentAction);
    connect(newTournamentAction, &QAction::triggered, this, &MainWindow::startNewTournament);

    QAction *loadTournamentAction = new QAction(tr("Turnier öffnen"), this);
    fileMenu->addAction(loadTournamentAction);
    connect(loadTournamentAction, &QAction::triggered, this, &MainWindow::triggerLoadTournament);

    m_closeTournamentAction = new QAction(tr("Turnier schließen"), this);
    fileMenu->addAction(m_closeTournamentAction);
    m_closeTournamentAction->setEnabled(false);
    connect(m_closeTournamentAction, &QAction::triggered, this, &MainWindow::closeTournament);

    fileMenu->addSeparator();

    m_settingsAction = new QAction(tr("Einstellungen"), this);
    connect(m_settingsAction, &QAction::triggered, this, &MainWindow::showSettingsDialog);
    fileMenu->addAction(m_settingsAction);
    m_settingsAction->setEnabled(false);

    m_exportAction = new QAction(tr("Turnierdaten exportieren"), this);
    connect(m_rankingPage, &RankingPage::enableExport, m_exportAction, &QAction::setEnabled);
    connect(m_exportAction, &QAction::triggered, this, &MainWindow::showExportDialog);
    fileMenu->addAction(m_exportAction);
    m_exportAction->setEnabled(false);

    fileMenu->addSeparator();

    QAction *quitAction = new QAction(tr("Beenden"), this);
    fileMenu->addAction(quitAction);
    connect(quitAction, &QAction::triggered, this, &MainWindow::quit);

    QMenu *extrasMenu = menuBar()->addMenu(tr("Extras"));

    QAction *stopWatchAction = new QAction(tr("Stoppuhr"), this);
    extrasMenu->addAction(stopWatchAction);
    connect(stopWatchAction, &QAction::triggered, this, &MainWindow::showStopWatch);

    QMenu *helpMenu = menuBar()->addMenu(tr("Hilfe"));

    QAction *handbookAction = new QAction(tr("Handbuch zu Muckturnier"), this);
    helpMenu->addAction(handbookAction);
    connect(handbookAction, &QAction::triggered, this, &MainWindow::showHandbook);

    QAction *aboutAction = new QAction(tr("Über Muckturnier"), this);
    helpMenu->addAction(aboutAction);
    connect(aboutAction, &QAction::triggered, this, &MainWindow::showAboutWindow);

    updateStatus(tr("Bitte ein neues Turnier starten oder ein bestehendes öffnen!"));
}

void MainWindow::startNewTournament()
{
    QString dbFileName = QFileDialog::getSaveFileName(this,
        tr("Bitte einen Dateinamen für die Turnierdatenbank auswählen"),
        QDir::toNativeSeparators(QStandardPaths::writableLocation(QStandardPaths::DocumentsLocation)
                                 + QString::fromUtf8("/Muckturnier.mtdb")),
        tr("Muckturnier-Datenbanken") + QString::fromUtf8(" (*.mtdb);;")
        + tr("Alle Dateien") + QString::fromUtf8(" (*)"));

    if (dbFileName.isEmpty()) {
        return;
    }

    // Yes, we have to do it in this way, because QFileInfo::isWritable
    // does not work as expected on Windows (returns false for "My Documents").
    bool canWrite;
    QFile test(dbFileName);
    if (test.open(QIODevice::WriteOnly)) {
        canWrite = true;
        test.close();
    } else {
        canWrite = false;
    }

    if (! canWrite) {
        QMessageBox::critical(this,
            tr("Neues Turnier starten"),
            tr("Es bestehen keine Schreibrechte für die Datei „%1“. "
               "Bitte die Zugriffsrechte überprüfen!").arg(QDir::toNativeSeparators(dbFileName)),
            QMessageBox::Ok);
        return;
    }

    QFileInfo dbFileInfo = QFileInfo(dbFileName);
    if (dbFileInfo.exists()) {
        // When we're overwriting an existing database, it has to be deleted
        dbFileInfo.dir().remove(dbFileInfo.fileName());
    }

    updateStatus(tr("Lege neue Turnierdatenbank an …"));
    QApplication::setOverrideCursor(Qt::WaitCursor);

    if (! m_db->create(dbFileName)) {
        // This should not happen
        updateStatus(tr("Das Anlegen der Turnierdatenbank „%1“ ist fehlgeschlagen!").arg(
            QDir::toNativeSeparators(dbFileName)));
        QApplication::restoreOverrideCursor();
        return;
    }

    databaseChanged(dbFileName);
    dbReset();
    m_settingsDialog->resetGui();

    QApplication::restoreOverrideCursor();
    showSettingsDialog();

    if (m_db->getDbFile().isEmpty()) {
        updateStatus(tr("Das Anlegen der neuen Turnierdatenbank wurde abgebrochen."));
    } else {
        updateStatus(tr("Turnier „%1“ gestartet.").arg(QDir::toNativeSeparators(dbFileName)));
    }
}

void MainWindow::triggerLoadTournament()
{
    loadTournament();
}

void MainWindow::loadTournament(QString dbToOpen)
{
    QString dbFileName;
    if (! dbToOpen.isEmpty()) {
        dbFileName = dbToOpen;
    } else {
        dbFileName = QFileDialog::getOpenFileName(this,
            tr("Bitte eine Turnierdatenbank auswählen"),
            QDir::toNativeSeparators(
                QStandardPaths::writableLocation(QStandardPaths::DocumentsLocation)),
            tr("Muckturnier-Datenbanken") + QString::fromUtf8(" (*.mtdb *.db);;")
            + tr("Alle Dateien") + QString::fromUtf8(" (*)"));
        if (dbFileName.isEmpty()) {
            return;
        }
    }

    QFileInfo fileInfo = QFileInfo(dbFileName);

    if (! fileInfo.exists()) {
        QMessageBox::critical(this,
            tr("Turnier öffnen"),
            tr("Das Öffnen der Turnierdatenbank „%1“ ist nicht möglich:<br/>"
               "Die Datei konnte nicht gefunden werden.").arg(
               QDir::toNativeSeparators(dbFileName)),
            QMessageBox::Ok);
        return;
    }

    if (! (fileInfo.isFile() || fileInfo.isSymLink())) {
        QMessageBox::critical(this,
            tr("Turnier öffnen"),
            tr("Das Öffnen der Turnierdatenbank „%1“ ist nicht möglich:<br/>"
               "Es ist weder eine Datei, noch ein symbolischer Link.").arg(
               QDir::toNativeSeparators(dbFileName)),
            QMessageBox::Ok);
        return;
    }

    if (! fileInfo.isReadable()) {
        QMessageBox::critical(this,
            tr("Turnier öffnen"),
            tr("Das Öffnen der Turnierdatenbank „%1“ ist nicht möglich:<br/>"
               "Die Datei ist nicht lesbar. "
               "Bitte die Zugriffsrechte überprüfen!").arg(
               QDir::toNativeSeparators(dbFileName)),
            QMessageBox::Ok);
        return;
    }

    if (! fileInfo.isWritable()) {
        QMessageBox::critical(this,
            tr("Turnier öffnen"),
            tr("Das Laden der Turnierdatenbank „%1“ ist nicht möglich:<br/>"
               "Die Datei ist nicht beschreibbar. "
               "Bitte die Zugriffsrechte überprüfen!").arg(
               QDir::toNativeSeparators(dbFileName)),
            QMessageBox::Ok);
        return;
    }

    // At least, we can read and write the file. So let's try to open it

    updateStatus(tr("Öffne Turnierdatenbank …"));
    QApplication::setOverrideCursor(Qt::WaitCursor);

    if (! m_db->openDb(dbFileName)) {
        QApplication::restoreOverrideCursor();
        QMessageBox::critical(this,
            tr("Turnier öffnen"),
            tr("<p>Das Öffnen der Turnierdatenbank „%1“ ist nicht möglich:</p><p>%2</p>").arg(
                QDir::toNativeSeparators(dbFileName), m_db->openError()),
            QMessageBox::Ok);
        return;
    }

    databaseChanged(dbFileName);

    if (m_db->tournamentStarted()) {
        m_settingsAction->setEnabled(true);
        m_settingsDialog->setToSaved();
        m_scoreDock->raise();
    } else {
        // The score page could be enabled by the previously used database,
        // so disable it first
        m_scoreDock->hide();
        // Re-enable the score page if the database already contains
        // the necessary number of players/pairs
        m_playersPage->reload();
        m_settingsAction->setEnabled(true);
        m_settingsDialog->setToSaved();
        m_playersDock->raise();
    }

    updateStatus(tr("Turnierdatenbank %1 geöffnet").arg(QDir::toNativeSeparators(dbFileName)));
    QApplication::restoreOverrideCursor();
}

void MainWindow::quit()
{
    QApplication::quit();
}

void MainWindow::showHandbook()
{
    // Opens the default browser to display the handbook.

    QString readmePath = resourcePath(QString::fromUtf8("Readme.htm"));
    if (! readmePath.isEmpty()) {
        QDesktopServices::openUrl(QUrl::fromLocalFile(readmePath));
        return;
    }

    QMessageBox::critical(this,
                          tr("Handbuch anzeigen"),
                          tr("<p>Das Handbuch kann nicht angezeigt werden: Die Datei „Readme.htm“ "
                             "wurde nicht gefunden.</p>"
                             "<p>Online ist unter <a href=\"http://nasauber.de/opensource/"
                             "muckturnier/readme.php\">http://nasauber.de/opensource/muckturnier/"
                             "readme.php</a> das aktuelle Handbuch zu finden.</p>"),
                          QMessageBox::Ok);
}

void MainWindow::showAboutWindow()
{
    QDialog *aboutWindow = new AboutWindow(this);
    aboutWindow->show();
}

void MainWindow::updateStatus(QString message)
{
    statusBar()->showMessage(message, 5000);
}

void MainWindow::databaseChanged(QString dbFileName)
{
    setWindowTitle(tr("Muckturnier: %1").arg(QFileInfo(dbFileName).fileName()));
    m_centralWidget->hide();
    m_settingsDialog->reload();
    setPlayersTabName();
    m_playersDock->show();
    m_playersPage->reload();

    m_exportAction->setEnabled(false);
    m_closeTournamentAction->setEnabled(true);

    checkEnableRankingPage();
    if (m_db->tournamentStarted()) {
        m_rankingPage->reload();
    }
}

void MainWindow::dbReset()
{
    setPlayersTabName();
    m_playersDock->hide();
    m_playersPage->reset();
    m_scoreDock->hide();
    m_rankingDock->hide();
}

void MainWindow::settingsSaved()
{
    setPlayersTabName();
    m_playersDock->show();
    m_playersPage->reload();
    m_playersDock->raise();
    m_settingsAction->setEnabled(true);
}

void MainWindow::setScoreTabEnabled(bool state)
{
    m_scoreDock->setVisible(state);
    if(state == true) {
        m_scorePage->reload();
    }
}

void MainWindow::tournamentStarted()
{
    setPlayersTabName();
    m_playersPage->reload();
    checkEnableRankingPage();
}

void MainWindow::setPlayersTabName()
{
    if (m_db->tournamentStarted()) {
        m_playersDock->setWindowTitle(m_db->playersString(Database::Numerus::plural));
    } else {
        m_playersDock->setWindowTitle(m_db->playersString(Database::Numerus::plural));
    }
}

void MainWindow::checkEnableRankingPage()
{
    m_rankingDock->setVisible(m_db->tournamentStarted());
}

void MainWindow::resizeEvent(QResizeEvent *event)
{
   QMainWindow::resizeEvent(event);
   if (m_scoreDock->isVisible()) {
        m_scorePage->setScrollAreaHeight();
   }
}

void MainWindow::showExportDialog()
{
    ExportDialog *exportDialog = new ExportDialog(this, m_db, m_rankingPage);
    connect(exportDialog, &ExportDialog::statusUpdate, this, &MainWindow::updateStatus);
    exportDialog->exec();
}

void MainWindow::showStopWatch()
{
    StopWatch *stopWatch = new StopWatch(this);
    stopWatch->show();
}

void MainWindow::showSettingsDialog()
{
    m_settingsDialog->updateDatabaseInfo();
    m_settingsDialog->exec();
}

void MainWindow::formatDockWidget(QDockWidget *dock)
{
    dock->setContextMenuPolicy(Qt::PreventContextMenu);
    dock->setFeatures(QDockWidget::DockWidgetMovable | QDockWidget::DockWidgetFloatable);
    dock->setStyleSheet(QString::fromUtf8("QDockWidget { font-weight: bold; font-size: %1pt }").arg(
        (int) double(dock->font().pointSize()) * 1.2));
}

QString MainWindow::resourcePath(QString fileName)
{
    if (m_possibleDataLocations.size() == 0) {
        // Get a list of possible locations for our external files
        QStringList possibleDataLocations;
        const QString iconsPath         = QString::fromUtf8("icons/hicolor/scalable/apps/");
        const QString resourcesPath     = QString::fromUtf8("muckturnier/");
        const QString documentationPath = QString::fromUtf8("doc/muckturnier/");

        // Add the binary's directory (possibly a build directory)
        possibleDataLocations << QCoreApplication::applicationDirPath() + QString::fromUtf8("/");

        // Add a possibly prefixed directory tree
        QUrl prefixed = QUrl(QCoreApplication::applicationDirPath() + QString::fromUtf8("/"))
                            .resolved(QString::fromUtf8("../share/"));
        possibleDataLocations << prefixed.toString() + iconsPath
                              << prefixed.toString() + resourcesPath
                              << prefixed.toString() + documentationPath;

        // Add system-wide locations
        for (const QString &path
            : QStandardPaths::standardLocations(QStandardPaths::GenericDataLocation)) {
            possibleDataLocations << path + QString::fromUtf8("/") + iconsPath
                                  << path + QString::fromUtf8("/") + resourcesPath
                                  << path + QString::fromUtf8("/") + documentationPath;
        }

        // Reverse the list
        while (possibleDataLocations.size() > 0) {
            const QString path = QDir::toNativeSeparators(possibleDataLocations.takeLast());
            if (! m_possibleDataLocations.contains(path)) {
                m_possibleDataLocations << path;
            }
        }
    }

    // Check if we already have it
    if (m_resourcePath.contains(fileName)) {
        return m_resourcePath[fileName];
    }

    // First, we search for embedded resources ...
    QDirIterator it(QString::fromUtf8(":"), QDirIterator::Subdirectories);
    while (it.hasNext()) {
        if (it.next() == QString::fromUtf8(":/") + fileName) {
            m_resourcePath[fileName] = QString::fromUtf8(":/") + fileName;
            return m_resourcePath[fileName];
        }
    }

    // ... then we check all other possible data locations
    for (const QString &path : m_possibleDataLocations) {
        QString filePath = path + QDir::separator() + fileName;
        if (QFileInfo::exists(filePath)) {
            m_resourcePath[fileName] = filePath;
            return m_resourcePath[fileName];
        }
    }

    // This will also return QString() if nothing is found
    return m_resourcePath[fileName];
}

void MainWindow::closeTournament()
{
    m_db->closeDatabaseConnection();
    m_playersDock->hide();
    m_scoreDock->hide();
    m_rankingDock->hide();
    m_centralWidget->show();
    m_closeTournamentAction->setEnabled(false);
    setWindowTitle(tr("Muckturnier"));
    updateStatus(tr("Bitte ein neues Turnier starten oder ein bestehendes öffnen!"));
}
