/* Copyright (c) 2012-2017 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "RankingPage.h"
#include "Database.h"
#include "Cell.h"
#include "version.h"

// Qt includes
#include <QDebug>
#include <QGroupBox>
#include <QVBoxLayout>
#include <QLabel>
#include <QTableWidget>
#include <QHeaderView>

RankingPage::RankingPage(QWidget *parent, Database *db)
    : QWidget(parent), m_db(db)
{
    QVBoxLayout *layout = new QVBoxLayout(this);

    m_rankingBox = new QGroupBox;
    QVBoxLayout *rankingBoxLayout = new QVBoxLayout(m_rankingBox);
    layout->addWidget(m_rankingBox);

    m_interimLabel = new QLabel;
    m_interimLabel->setWordWrap(true);
    rankingBoxLayout->addWidget(m_interimLabel);
    m_interimLabel->hide();

    m_ranking = new QTableWidget;
    m_ranking->verticalHeader()->hide();
    m_ranking->setShowGrid(false);
    m_ranking->setWordWrap(false);
    m_ranking->setSelectionMode(QTableWidget::NoSelection);
    m_ranking->setFocusPolicy(Qt::NoFocus);
    m_ranking->setColumnCount(4);
    rankingBoxLayout->addWidget(m_ranking);
}

void RankingPage::reload()
{
    QMap<int, bool> roundsState = m_db->roundsState();
    m_allRounds = roundsState.size();

    QList<int> incompleteRounds;
    QList<int> roundNumbers = roundsState.keys();
    qSort(roundNumbers);
    for (const int &round : roundNumbers) {
        if (! roundsState[round]) {
            incompleteRounds << round;
        }
    }

    int missingRounds = incompleteRounds.size();
    if (missingRounds != 0) {
        emit enableExport(false);
    } else {
        emit enableExport(true);
    }

    if (missingRounds > 0) {
        m_rankingBox->setTitle(tr("Rangliste im Verlauf der %1. Runde").arg(
                                  QString::number(roundsState.size())));

        QString interimText = tr("<p><b>Achtung!</b> Das ist nur ein Zwischenergebnis. "
                                 "Es fehlen noch Ergebnisse aus ");
        if (missingRounds == 1) {
            interimText += tr("der Runde %1.").arg(QString::number(incompleteRounds[0]));
        } else if (missingRounds == 2) {
            interimText += tr("den Runden %1 und %2.").arg(QString::number(incompleteRounds[0]),
                                                           QString::number(incompleteRounds[1]));
        } else {
            interimText += tr("den Runden ");
            for (int i = 0; i < incompleteRounds.size() - 1; i++) {
                interimText += tr("%1, ").arg(QString::number(incompleteRounds[i]));
            }
            interimText = interimText.left(interimText.size() - 2);
            interimText += tr(" und %1.").arg(QString::number(incompleteRounds.last()));
        }
        interimText += tr("</p>"
                          "<p>Paare, von denen noch Spielstände fehlen, sind "
                          "<span style=\"color:red\">rot</span> markiert.</p>");

        m_interimLabel->setText(interimText);
        m_interimLabel->show();
    } else {
        m_interimLabel->hide();
        if (m_allRounds == 1) {
            m_rankingBox->setTitle(tr("Rangliste nach einer Runde"));
        } else {
            m_rankingBox->setTitle(tr("Rangliste nach %1 Runden").arg(
                                      QString::number(m_allRounds)));
        }
    }

    QList<Database::Ranking> allRankings = m_db->getRanking();
    int allGames = roundsState.size() * m_db->getGamesPerRound();

    m_ranking->clearContents();
    m_ranking->setRowCount(0);
    m_ranking->setHorizontalHeaderLabels(QStringList()
        << tr("Platz")
        << m_db->playersString(Database::Numerus::singular)
        << tr("Bobbl")
        << tr("Tore"));

    m_ranking->setRowCount(allRankings.size());
    int rank = 0;
    int trueRank = 0;
    int prevRank = 0;
    int prevBoogers = 0;
    int prevGoals = 0;

    int color = 1;
    for (const Database::Ranking &ranking : allRankings) {
        trueRank++;
        if (ranking.boogers != prevBoogers || ranking.goals != prevGoals) {
            rank = trueRank;
        } else {
            rank = prevRank;
        }

        if (rank != prevRank) {
            color++;
            color = color % 2;
        }

        prevRank = rank;
        prevBoogers = ranking.boogers;
        prevGoals = ranking.goals;

        Qt::GlobalColor textColor = Qt::color0;
        if (ranking.playedGames != allGames) {
            textColor = Qt::red;
        }

        m_ranking->setItem(
            trueRank - 1, 0, new Cell(QString::number(rank), false, color, textColor));
        m_ranking->setItem(
            trueRank - 1, 1, new Cell(ranking.name, false, color, textColor));
        m_ranking->setItem(
            trueRank - 1, 2, new Cell(QString::number(ranking.boogers), false, color, textColor));
        m_ranking->setItem(
            trueRank - 1, 3, new Cell(QString::number(ranking.goals), false, color, textColor));
    }

    m_ranking->resizeColumnsToContents();
}
