/* Copyright (c) 2012-2017 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "SettingsDialog.h"
#include "Database.h"
#include "MainWindow.h"

// Qt includes
#include <QDebug>
#include <QVBoxLayout>
#include <QGroupBox>
#include <QGridLayout>
#include <QLabel>
#include <QRadioButton>
#include <QSpinBox>
#include <QPushButton>
#include <QMessageBox>
#include <QApplication>
#include <QDir>
#include <QCloseEvent>
#include <QDialogButtonBox>

SettingsDialog::SettingsDialog(QWidget *parent, Database *db)
    : QDialog(parent), m_db(db)
{
    QVBoxLayout *layout = new QVBoxLayout(this);
    layout->setSizeConstraint(QLayout::SetFixedSize);

    // Database

    m_databaseBox = new QGroupBox(tr("Datenbank"));
    QGridLayout *databaseBoxLayout = new QGridLayout(m_databaseBox);
    layout->addWidget(m_databaseBox);

    QLabel *dbFileHeaderLabel = new QLabel(tr("Turnierdatenbank:"));
    dbFileHeaderLabel->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    databaseBoxLayout->addWidget(dbFileHeaderLabel, 0, 0);
    m_dbFileLabel = new QLabel;
    m_dbFileLabel->setTextInteractionFlags(Qt::TextSelectableByMouse);
    databaseBoxLayout->addWidget(m_dbFileLabel, 0, 1);

    m_mtVersionHeaderLabel = new QLabel(tr("Erstellt mit:"));
    m_mtVersionHeaderLabel->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    databaseBoxLayout->addWidget(m_mtVersionHeaderLabel, 1, 0);
    m_mtVersionLabel = new QLabel;
    m_mtVersionLabel->setTextInteractionFlags(Qt::TextSelectableByMouse);
    databaseBoxLayout->addWidget(m_mtVersionLabel, 1, 1);

    // Tournament type

    QGroupBox *tournamentTypeBox = new QGroupBox(tr("Turniertyp"));
    QGridLayout *tournamentTypeLayout = new QGridLayout(tournamentTypeBox);
    layout->addWidget(tournamentTypeBox);

    m_fixedPairsLabel = new QLabel(tr("Feste Paare:"));
    m_fixedPairsLabel->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    m_fixedPairsLabel->hide();
    tournamentTypeLayout->addWidget(m_fixedPairsLabel, 0, 0, Qt::AlignTop);

    m_pairs = new QRadioButton(tr("Feste Paare"));
    m_pairs->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    tournamentTypeLayout->addWidget(m_pairs, 0, 1, Qt::AlignTop);

    m_fixedPairsDescription = new QLabel(tr(
        "<p>"
        "Es werden feste Paare angemeldet, die sich nicht ändern. Entweder wird pro Runde "
        "ausgelost, welche Paare gegeneinander spielen, oder es wird nach folgendem festen Schema "
        "gespielt: Jedes Paar lost einen Start-Tisch und die Rolle „Paar 1“ bzw. „Paar 2“. Paar 1 "
        "bleibt sitzen, Paar 2 rückt jede Runde einen Tisch weiter."
        "</p>"
    ));
    m_fixedPairsDescription->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Minimum);
    m_fixedPairsDescription->setWordWrap(true);
    tournamentTypeLayout->addWidget(m_fixedPairsDescription, 0, 2, Qt::AlignTop);

    m_singlePlayersLabel = new QLabel(tr("Einzelne Spieler:"));
    m_singlePlayersLabel->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    m_singlePlayersLabel->hide();
    tournamentTypeLayout->addWidget(m_singlePlayersLabel, 1, 0, Qt::AlignTop);

    m_singlePlayers = new QRadioButton(tr("Einzelne Spieler"));
    m_singlePlayers->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    tournamentTypeLayout->addWidget(m_singlePlayers, 1, 1, Qt::AlignTop);

    m_singlePlayersDescription = new QLabel(tr(
        "<p>"
        "Es werden einzelne Spieler angemeldet. In jeder Runde werden die Spieler, die in einem "
        "Paar zusammenspielen, und die Paare, die gegeneinander spielen, neu ausgelost."
        "</p>"
    ));
    m_singlePlayersDescription->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Minimum);
    m_singlePlayersDescription->setWordWrap(true);
    tournamentTypeLayout->addWidget(m_singlePlayersDescription, 1, 2, Qt::AlignTop);

    // Game settings

    QGroupBox *gameSettingsBox = new QGroupBox(tr("Spieleinstellungen"));
    QHBoxLayout *gameSettingsMainLayout = new QHBoxLayout(gameSettingsBox);
    QGridLayout *gameSettingsLayout = new QGridLayout;
    gameSettingsMainLayout->addLayout(gameSettingsLayout);
    layout->addWidget(gameSettingsBox);

    QLabel *pointsPerBooger = new QLabel(tr("Punkte pro Bobbl:"));
    pointsPerBooger->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    gameSettingsLayout->addWidget(pointsPerBooger, 0, 0);

    m_boogerScore = new QSpinBox;
    m_boogerScore->setMinimum(1);
    m_boogerScore->setMaximum(999);
    m_boogerScore->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    gameSettingsLayout->addWidget(m_boogerScore, 0, 1);

    m_boogerScoreLabel = new QLabel;
    gameSettingsLayout->addWidget(m_boogerScoreLabel, 0, 2);

    QLabel *boogersPerRound = new QLabel(tr("Bobbl pro Runde:"));
    boogersPerRound->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    gameSettingsLayout->addWidget(boogersPerRound, 1, 0);

    m_gamesPerRound = new QSpinBox;
    m_gamesPerRound->setMinimum(1);
    m_gamesPerRound->setMaximum(999);
    m_gamesPerRound->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    gameSettingsLayout->addWidget(m_gamesPerRound, 1, 1);

    m_gamesPerRoundLabel = new QLabel;
    gameSettingsLayout->addWidget(m_gamesPerRoundLabel, 1, 2);

    gameSettingsMainLayout->addStretch();

    // Save tournament settings

    m_buttonBox = new QDialogButtonBox(QDialogButtonBox::Ok
                                       | QDialogButtonBox::Cancel
                                       | QDialogButtonBox::Reset);

    m_buttonBox->button(QDialogButtonBox::Ok)->setText(tr("Turniereinstellungen speichern"));
    m_buttonBox->button(QDialogButtonBox::Cancel)->setText(tr("Abbrechen"));
    m_buttonBox->button(QDialogButtonBox::Reset)->setText(tr("Datenbank zurücksetzen"));

    connect(m_buttonBox->button(QDialogButtonBox::Ok), &QPushButton::clicked,
            this, &SettingsDialog::saveSettings);
    connect(m_buttonBox->button(QDialogButtonBox::Cancel), &QPushButton::clicked,
            this, &QDialog::close);
    connect(m_buttonBox->button(QDialogButtonBox::Reset), &QPushButton::clicked,
            this, &SettingsDialog::resetDatabase);

    layout->addWidget(m_buttonBox);

    resetGui();
}

void SettingsDialog::resetGui()
{
    m_buttonBox->button(QDialogButtonBox::Reset)->hide();
    m_buttonBox->button(QDialogButtonBox::Ok)->show();
    m_buttonBox->button(QDialogButtonBox::Cancel)->setText(tr("Abbrechen"));
    m_fixedPairsLabel->hide();
    m_fixedPairsDescription->show();
    m_singlePlayersLabel->hide();
    m_singlePlayersDescription->show();
    m_pairs->show();
    m_singlePlayers->show();
    m_boogerScoreLabel->hide();
    m_boogerScore->show();
    m_gamesPerRoundLabel->hide();
    m_gamesPerRound->show();
    m_buttonBox->button(QDialogButtonBox::Ok)->setFocus();
}

void SettingsDialog::resetDatabase()
{
    QMessageBox messageBox(this);
    messageBox.setIcon(QMessageBox::Question);
    messageBox.setWindowTitle(tr("Datenbank zurücksetzen"));
    messageBox.setText(tr("Sollen wirklich alle Einstellungen und Ergebnisse gelöscht werden?"));
    messageBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
    messageBox.setButtonText(QMessageBox::Yes, tr("Ja"));
    messageBox.setButtonText(QMessageBox::No, tr("Nein"));
    messageBox.setDefaultButton(QMessageBox::No);

    if (messageBox.exec() == QMessageBox::No) {
        return;
    }

    emit statusUpdate(tr("Setzte Turnierdatenbank zurück …"));
    QApplication::setOverrideCursor(Qt::WaitCursor);

    m_db->reset();
    resetGui();
    reload();
    emit dbReset();

    QApplication::restoreOverrideCursor();
    emit statusUpdate(tr("Turnierdatenbank zurückgesetzt"));
}

void SettingsDialog::saveSettings()
{
    emit statusUpdate(tr("Speichere Turniereinstellungen …"));
    QApplication::setOverrideCursor(Qt::WaitCursor);

    if (m_pairs->isChecked()) {
        m_db->setTournamentMode(Database::TournamentMode::fixedPairs);
    } else {
        m_db->setTournamentMode(Database::TournamentMode::singlePlayers);
    }
    m_db->setBoogerScore(m_boogerScore->value());
    m_db->setGamesPerRound(m_gamesPerRound->value());

    m_db->saveSettings();
    setToSaved();
    emit settingsSaved();
    QApplication::restoreOverrideCursor();
    emit statusUpdate(tr("Turniereinstellungen gespeichert"));

    accept();
}

void SettingsDialog::reload()
{
    switch (m_db->getTournamentMode()) {
    case Database::TournamentMode::fixedPairs:
        m_pairs->setChecked(Qt::Checked);
        m_singlePlayers->setChecked(Qt::Unchecked);
        break;
    case Database::TournamentMode::singlePlayers:
        m_pairs->setChecked(Qt::Unchecked);
        m_singlePlayers->setChecked(Qt::Checked);
        break;
    }

    m_boogerScore->setValue(m_db->getBoogerScore());
    m_gamesPerRound->setValue(m_db->getGamesPerRound());
}

void SettingsDialog::setToSaved()
{
    m_buttonBox->button(QDialogButtonBox::Reset)->show();
    m_buttonBox->button(QDialogButtonBox::Ok)->hide();
    m_buttonBox->button(QDialogButtonBox::Cancel)->setText(tr("Schließen"));

    m_boogerScoreLabel->setText(QString::number(m_boogerScore->value()));
    m_boogerScoreLabel->show();
    m_boogerScore->hide();
    m_gamesPerRoundLabel->setText(QString::number(m_gamesPerRound->value()));
    m_gamesPerRoundLabel->show();
    m_gamesPerRound->hide();

    switch (m_db->getTournamentMode()) {
    case Database::TournamentMode::fixedPairs:
        m_fixedPairsLabel->show();
        m_fixedPairsDescription->show();
        m_singlePlayersLabel->hide();
        m_singlePlayersDescription->hide();
        break;
    case Database::TournamentMode::singlePlayers:
        m_fixedPairsLabel->hide();
        m_fixedPairsDescription->hide();
        m_singlePlayersLabel->show();
        m_singlePlayersDescription->show();
        break;
    }
    m_pairs->hide();
    m_singlePlayers->hide();

    m_buttonBox->button(QDialogButtonBox::Cancel)->setFocus();
}

void SettingsDialog::updateDatabaseInfo()
{
    m_dbFileLabel->setText(QDir::toNativeSeparators(m_db->getDbFile()));
    m_mtVersionLabel->setText(tr("Muckturnier %1").arg(m_db->getMuckturnierVersion()));
}

void SettingsDialog::closeEvent(QCloseEvent *event)
{
    if (m_buttonBox->button(QDialogButtonBox::Ok)->isVisible()) {
        const QString dbFile = m_db->getDbFile();

        QMessageBox messageBox(this);
        messageBox.setIcon(QMessageBox::Question);
        messageBox.setWindowTitle(tr("Starten des neuen Turniers abbrechen"));
        messageBox.setText(tr("Soll die bereits angelegte (leere) Turnierdatenbank\n"
                              "%1\n"
                              "gelöscht werden?").arg(QDir::toNativeSeparators(dbFile)));
        messageBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No | QMessageBox::Cancel);
        messageBox.setButtonText(QMessageBox::Yes, tr("Ja"));
        messageBox.setButtonText(QMessageBox::No, tr("Nein"));
        messageBox.setButtonText(QMessageBox::Cancel, tr("Abbrechen"));
        messageBox.setDefaultButton(QMessageBox::Yes);

        int reply = messageBox.exec();
        if (reply == QMessageBox::Cancel) {
            event->ignore();
            return;
        }
        if (reply == QMessageBox::Yes) {
            m_db->closeDatabaseConnection();
            QDir dir;
            if (! dir.remove(dbFile)) {
                QMessageBox::warning(this,
                    tr("Starten des neuen Turniers abbrechen"),
                    tr("Konnte die Datenbank nicht löschen!"));
            }
        }
        dynamic_cast<MainWindow*>(parent())->closeTournament();
    }
    QDialog::reject();
}

void SettingsDialog::reject()
{
    if (m_buttonBox->button(QDialogButtonBox::Ok)->isVisible()) {
        // Pressing ESC behaves like pressing "Cancel"
        QDialog::close();
    } else {
        // Pressing ESC closes the dialog
        QDialog::reject();
    }
}
